#!/usr/bin/env python

# Post-processing functions that run after the main parsing.

from . import re
from .extras import link_patterns
from .patterns import episode_name_pattern, langs, patterns, pre_website_encoder_pattern

# Before excess functions (before we split what was unmatched in the title into a list).
# They all take in the parse object and what was unmatched, and must return the latter minus
# what they used.


# Try and find the episode name.
def try_episode_name(self, unmatched):
    match = re.findall(episode_name_pattern, unmatched)
    # First we see if there's a match in unmatched, then we look if it's after an episode
    # or a day in the full release title.
    if match:
        match = re.search(
            "(?:"
            + link_patterns(patterns["episode"])
            + "|"
            + patterns["day"]
            + r")[._\-\s+]*("
            + re.escape(match[0])
            + ")",
            self.torrent_name,
            re.IGNORECASE,
        )
        if match:
            match_s, match_e = match.start(len(match.groups())), match.end(
                len(match.groups())
            )
            match = match.groups()[-1]
            self._part("episodeName", (match_s, match_e), self._clean_string(match))
            unmatched = unmatched.replace(match, "")
    return unmatched


def try_encoder_before_site(self, unmatched):
    match = re.findall(pre_website_encoder_pattern, unmatched.strip())

    if match:
        found_match = None
        for m in match:
            full_title_match = re.search(
                r"[\s\-]("
                + re.escape(m)
                + ")(?:\."
                + link_patterns(patterns["filetype"])
                + ")?$",
                self.torrent_name,
                re.I,
            )
            if full_title_match:
                found_match = full_title_match
                break
        match = found_match
        if match:
            match_s, match_e = match.start(0), match.end(0)
            encoder_and_site = list(
                filter(None, re.split(r"[\-\s\)]", match.groups()[0]))
            )
            if len(encoder_and_site) == 2:
                encoder_raw = encoder_and_site[0]
                site_raw = encoder_and_site[1]
                self._part(
                    "encoder",
                    (match_s, match_e - len(site_raw)),
                    self._clean_string(encoder_raw),
                )
                self._part(
                    "site",
                    (match_s + len(encoder_raw), match_e),
                    self._clean_string(site_raw),
                    overwrite=True,
                )
                unmatched = unmatched.replace(match.group(0), "")

    return unmatched


post_processing_before_excess = [
    try_episode_name,
    try_encoder_before_site,
]


# After excess functions take in just the parse object, and shouldn't return anything.


# encoder is assumed to be the last element of `excess`, if not already added.
def try_encoder(self):
    if "excess" not in self.parts or "encoder" in self.parts:
        return
    excess = self.parts["excess"]
    if not isinstance(excess, list):
        excess = [excess]

    if excess:
        encoder = excess.pop()
        self._part("encoder", None, encoder, overwrite=True)

    if not excess:
        self.parts.pop("excess")
    else:
        self._part("excess", None, excess, overwrite=True)


# Split encoder name and site, adding the latter to self.parts
def try_site(self):
    if "encoder" not in self.parts or "website" in self.parts:
        return
    encoder = self.parts["encoder"]
    if self.coherent_types:
        encoder = encoder[0]
    pat = r"(\[(.*)\])"
    match = re.findall(pat, encoder, re.IGNORECASE)
    if match:
        match = match[0]
        raw = match[0]
        if match:
            if not re.match(r"[\[\],.+\-]*\Z", match[1], re.IGNORECASE):
                self._part("site", None, match[1])
            self._part("encoder", None, encoder.replace(raw, ""), overwrite=True)


# If this match starts like the language one did, the only match for language
# and subtitles is a list of langs directly followed by a subs-string. When this
# is true, they would both match on it, but what it likely means is that all the
# langs are language, and the subs string just indicates the existance of subtitles.
# (e.g. Ita.Eng.MSubs would match Ita and Eng for language and subs - this makes
# subs only become MSubs, and leaves language as Ita and Eng)
def fix_same_subtitles_language_match(self):
    if (
        "language" in self.part_slices
        and "subtitles" in self.part_slices
        and self.part_slices["language"][0] == self.part_slices["subtitles"][0]
    ):
        subs = self.parts["subtitles"][-1]
        if self.standardise:
            subs = "Available"
        self._part("subtitles", None, subs, overwrite=True)


# If there are no languages, but subtitles were matched, we should assume the first lang
# is the actual language, and remove it from the subtitles.
def fix_subtitles_no_language(self):
    if (
        "language" not in self.parts
        and "subtitles" in self.parts
        and isinstance(self.parts["subtitles"], list)
        and len(self.parts["subtitles"]) > 1
    ):
        self._part("language", None, self.parts["subtitles"][0])
        self._part("subtitles", None, self.parts["subtitles"][1:], overwrite=True)


# Language matches, to support multi-language releases that have the audio with each
# language, will contain audio info (or simply extra strings like 'dub').
# We remove non-lang matching items from this list.
def filter_non_languages(self):
    if "language" in self.parts and isinstance(self.parts["language"], list):
        languages = list(self.parts["language"])
        for lang in self.parts["language"]:
            matched = False
            for (lang_regex, lang_clean) in langs:
                if re.match(lang_regex, lang, re.IGNORECASE):
                    matched = True
                    break
            if not matched:
                languages.remove(lang)

        self._part("language", self.part_slices["language"], languages, overwrite=True)


def try_vague_season_episode(self):
    title = self.parts["title"]
    m = re.search("(\d{1,2})-(\d{1,2})$", title)
    if m:
        if "season" not in self.parts and "episode" not in self.parts:
            new_title = title[: m.start()]
            offset = self.part_slices["title"][0]
            # Setting the match slices here doesn't actually matter, but good practice.
            self._part(
                "season", (offset + m.start(1), offset + m.end(1)), int(m.group(1))
            )
            self._part(
                "episode", (offset + m.start(2), offset + m.end(2)), int(m.group(2))
            )
            self._part(
                "title",
                (offset, offset + len(new_title)),
                self._clean_string(new_title),
                overwrite=True,
            )


post_processing_after_excess = [
    try_encoder,
    try_site,
    fix_same_subtitles_language_match,
    fix_subtitles_no_language,
    filter_non_languages,
    try_vague_season_episode,
]



#!/usr/bin/env python
from . import re
from .extras import exceptions, genres, langs, link_patterns, patterns_ignore_title
from .patterns import delimiters, patterns, patterns_ordered, types
from .post import post_processing_after_excess, post_processing_before_excess


class PTN(object):
    def __init__(self):
        self.torrent_name = None
        self.parts = None
        self.part_slices = None
        self.match_slices = None
        self.standardise = None
        self.coherent_types = None

        self.post_title_pattern = "(?:{}|{})".format(
            link_patterns(patterns["season"]), link_patterns(patterns["year"])
        )

    def _part(self, name, match_slice, clean, overwrite=False):
        if overwrite or name not in self.parts:
            if self.coherent_types:
                if name not in ["title", "episodeName"] and not isinstance(clean, bool):
                    if not isinstance(clean, list):
                        clean = [clean]
            else:
                if isinstance(clean, list) and len(clean) == 1:
                    clean = clean[0]  # Avoids making a list if it only has 1 element

            self.parts[name] = clean
            self.part_slices[name] = match_slice

        # Ignored patterns will still be considered 'matched' to remove them from excess.
        if match_slice:
            self.match_slices.append(match_slice)

    @staticmethod
    def _clean_string(string):
        clean = re.sub(r"^( -|\(|\[)", "", string)
        if clean.find(" ") == -1 and clean.find(".") != -1:
            clean = re.sub(r"\.", " ", clean)
        clean = re.sub(r"_", " ", clean)
        clean = re.sub(r"([\[)_\]]|- )$", "", clean).strip()
        clean = clean.strip(" _-")

        return clean

    def parse(self, name, standardise, coherent_types):
        name = name.strip()
        self.parts = {}
        self.part_slices = {}
        self.torrent_name = name
        self.match_slices = []
        self.standardise = standardise
        self.coherent_types = coherent_types

        for key, pattern_options in [(key, patterns[key]) for key in patterns_ordered]:
            pattern_options = self.normalise_pattern_options(pattern_options)

            for (pattern, replace, transforms) in pattern_options:
                if key not in ("season", "episode", "site", "language", "genre"):
                    pattern = r"\b(?:{})\b".format(pattern)

                clean_name = re.sub(r"_", " ", self.torrent_name)
                matches = self.get_matches(pattern, clean_name, key)

                if not matches:
                    continue

                # With multiple matches, we will usually want to use the first match.
                # For 'year', we instead use the last instance of a year match since,
                # if a title includes a year, we don't want to use this for the year field.
                match_index = 0
                if key == "year":
                    match_index = -1

                match = matches[match_index]["match"]
                match_start, match_end = (
                    matches[match_index]["start"],
                    matches[match_index]["end"],
                )
                if (
                    key in self.parts
                ):  # We can skip ahead if we already have a matched part
                    self._part(key, (match_start, match_end), None, overwrite=False)
                    continue

                index = self.get_match_indexes(match)

                if key in ("season", "episode"):
                    clean = self.get_season_episode(match)
                elif key == "subtitles":
                    clean = self.get_subtitles(match)
                elif key in ("language", "genre"):
                    clean = self.split_multi(match)
                elif key in types.keys() and types[key] == "boolean":
                    clean = True
                else:
                    clean = match[index["clean"]]
                    if key in types.keys() and types[key] == "integer":
                        clean = int(clean)

                if self.standardise:
                    clean = self.standardise_clean(clean, key, replace, transforms)

                self._part(key, (match_start, match_end), clean)

        self.process_title()
        self.fix_known_exceptions()

        unmatched = self.get_unmatched()
        for f in post_processing_before_excess:
            unmatched = f(self, unmatched)

        # clean_unmatched() depends on the before_excess methods adding more match slices.
        cleaned_unmatched = self.clean_unmatched()
        if cleaned_unmatched:
            self._part("excess", None, cleaned_unmatched)

        for f in post_processing_after_excess:
            f(self)

        return self.parts

    # Handles all the optional/missing tuple elements into a consistent list.
    @staticmethod
    def normalise_pattern_options(pattern_options):
        pattern_options_norm = []

        if isinstance(pattern_options, tuple):
            pattern_options = [pattern_options]
        elif not isinstance(pattern_options, list):
            pattern_options = [(pattern_options, None, None)]
        for options in pattern_options:
            if len(options) == 2:  # No transformation
                pattern_options_norm.append(options + (None,))
            elif isinstance(options, tuple):
                if isinstance(options[2], tuple):
                    pattern_options_norm.append(
                        tuple(list(options[:2]) + [[options[2]]])
                    )
                elif isinstance(options[2], list):
                    pattern_options_norm.append(options)
                else:
                    pattern_options_norm.append(
                        tuple(list(options[:2]) + [[(options[2], [])]])
                    )

            else:
                pattern_options_norm.append((options, None, None))
        pattern_options = pattern_options_norm
        return pattern_options

    def get_matches(self, pattern, clean_name, key):
        grouped_matches = []
        matches = list(re.finditer(pattern, clean_name, re.IGNORECASE))
        for m in matches:
            if m.start() < self.ignore_before_index(clean_name, key):
                continue
            groups = m.groups()
            if not groups:
                grouped_matches.append((m.group(), m.start(), m.end()))
            else:
                grouped_matches.append((groups, m.start(), m.end()))

        parsed_matches = []
        for match in grouped_matches:
            m = match[0]
            if isinstance(m, tuple):
                m = list(m)
            else:
                m = [m]
            parsed_matches.append({"match": m, "start": match[1], "end": match[2]})
        return parsed_matches

    # Only use part of the torrent name after the (guessed) title (split at a season or year)
    # to avoid matching certain patterns that could show up in a release title.
    def ignore_before_index(self, clean_name, key):
        match = None
        if key in patterns_ignore_title:
            patterns_ignored = patterns_ignore_title[key]
            if not patterns_ignored:
                match = re.search(self.post_title_pattern, clean_name, re.IGNORECASE)
            else:
                for ignore_pattern in patterns_ignored:
                    if re.findall(ignore_pattern, clean_name, re.IGNORECASE):
                        match = re.search(
                            self.post_title_pattern, clean_name, re.IGNORECASE
                        )

        if match:
            return match.start()
        return 0

    @staticmethod
    def get_match_indexes(match):
        index = {"raw": 0, "clean": 0}

        if len(match) > 1:
            # for season we might have it in index 1 or index 2
            # e.g. "5x09" TODO is this weirdness necessary
            for i in range(1, len(match)):
                if match[i]:
                    index["clean"] = i
                    break

        return index

    @staticmethod
    def get_season_episode(match):
        clean = None
        m = re.findall(r"[0-9]+", match[0])
        if m and len(m) > 1:
            clean = list(range(int(m[0]), int(m[-1]) + 1))
        elif m:
            clean = int(m[0])
        return clean

    @staticmethod
    def split_multi(match):
        # handle multi languages
        m = re.split(r"{}+".format(delimiters), match[0])
        clean = list(filter(None, m))

        return clean

    @staticmethod
    def get_subtitles(match):
        # handle multi subtitles
        m = re.split(r"{}+".format(delimiters), match[0])
        m = list(filter(None, m))
        clean = []
        # If it's only 1 result, it's fine if it's just 'subs'.
        if len(m) == 1:
            clean = m
        else:
            for x in m:
                if not re.match("subs?|soft", x, re.I):
                    clean.append(x)

        return clean

    def standardise_clean(self, clean, key, replace, transforms):
        if replace:
            clean = replace
        if transforms:
            for transform in filter(lambda t: t[0], transforms):
                # For python2 compatibility, we're not able to simply pass functions as str.upper
                # means different things in 2.7 and 3.5.
                clean = getattr(clean, transform[0])(*transform[1])
        if key == "language" or key == "subtitles":
            clean = self.standardise_languages(clean)
            if not clean:
                clean = "Available"
        if key == "genre":
            clean = self.standardise_genres(clean)
        return clean

    @staticmethod
    def standardise_languages(clean):
        cleaned_langs = []
        for lang in clean:
            for (lang_regex, lang_clean) in langs:
                if re.match(
                    lang_regex,
                    re.sub(
                        link_patterns(patterns["subtitles"][2:]), "", lang, flags=re.I
                    ),
                    re.IGNORECASE,
                ):
                    cleaned_langs.append(lang_clean)
                    break
        clean = cleaned_langs
        return clean

    @staticmethod
    def standardise_genres(clean):
        standard_genres = []
        for genre in clean:
            for (regex, clean) in genres:
                if re.match(regex, genre, re.IGNORECASE):
                    standard_genres.append(clean)
                    break
        return standard_genres

    # Merge all the match slices (such as when they overlap), then remove
    # them from excess.
    def merge_match_slices(self):
        matches = sorted(self.match_slices, key=lambda match: match[0])

        i = 0
        slices = []
        while i < len(matches):
            start, end = matches[i]
            i += 1
            for (next_start, next_end) in matches[i:]:
                if next_start <= end:
                    end = max(end, next_end)
                    i += 1
                else:
                    break
            slices.append((start, end))

        self.match_slices = slices

    def process_title(self):
        unmatched = self.unmatched_list(keep_punctuation=False)

        # Use the first one as the title
        if unmatched:
            title_start, title_end = unmatched[0][0], unmatched[0][1]

            # If our unmatched is after the first 3 matches, we assume the title is missing
            # (or more likely got parsed as something else), as no torrents have it that
            # far away from the beginning of the release title.
            if (
                len(self.part_slices) > 3
                and title_start
                > sorted(self.part_slices.values(), key=lambda s: s[0])[3][0]
            ):
                self._part("title", None, "")

            raw = self.torrent_name[title_start:title_end]
            # Something in square brackets with 3 chars or less is too weird to be right.
            # If this seems too arbitrary, make it any square bracket, and Mother test
            # case will lose its translated title (which is mostly fine I think).
            m = re.search("\(|(?:\[(?:.{,3}\]|[^\]]*\d[^\]]*\]?))", raw, flags=re.I)
            if m:
                relative_title_end = m.start()
                raw = raw[:relative_title_end]
                title_end = relative_title_end + title_start
            # Similar logic as above, but looking at beginning of string unmatched brackets.
            m = re.search("^(?:\)|\[.*\])", raw)
            if m:
                relative_title_start = m.end()
                raw = raw[relative_title_start:]
                title_start = relative_title_start + title_start
            clean = self._clean_string(raw)
            # Re-add title_start to unrelative the index from raw to self.torrent_name
            self._part("title", (title_start, title_end), clean)
        else:
            self._part("title", None, "")

    def unmatched_list(self, keep_punctuation=True):
        self.merge_match_slices()
        unmatched = []
        prev_start = 0
        # A default so the last append won't crash if nothing has matched
        end = len(self.torrent_name)
        # Find all unmatched strings that aren't just punctuation
        for (start, end) in self.match_slices:
            if keep_punctuation or not re.match(
                delimiters + r"*\Z", self.torrent_name[prev_start:start]
            ):
                unmatched.append((prev_start, start))
            prev_start = end

        # Add the last unmatched slice
        if keep_punctuation or not re.match(
            delimiters + r"*\Z", self.torrent_name[end:]
        ):
            unmatched.append((end, len(self.torrent_name)))

        return unmatched

    def fix_known_exceptions(self):
        # Considerations for results that are known to cause issues, such
        # as media with years in them but without a release year.
        for exception in exceptions:
            incorrect_key, incorrect_value = exception["incorrect_parse"]
            if (
                self.parts["title"] == exception["parsed_title"]
                and incorrect_key in self.parts
            ):
                if self.parts[incorrect_key] == incorrect_value or (
                    self.coherent_types and incorrect_value in self.parts[incorrect_key]
                ):
                    self.parts.pop(incorrect_key)
                    self._part("title", None, exception["actual_title"], overwrite=True)

    def get_unmatched(self):
        unmatched = ""
        for (start, end) in self.unmatched_list():
            unmatched += self.torrent_name[start:end]

        return unmatched

    def clean_unmatched(self):
        unmatched = []
        for (start, end) in self.unmatched_list():
            unmatched.append(self.torrent_name[start:end])

        unmatched_clean = []
        for raw in unmatched:
            clean = re.sub(r"(^[-_.\s(),]+)|([-.\s,]+$)", "", raw)
            clean = re.sub(r"[()/]", " ", clean)
            unmatched_clean += re.split(r"\.\.+|\s+", clean)

        filtered = []
        for extra in unmatched_clean:
            # re.fullmatch() is not available in python 2.7, so we manually do it with \Z.
            if not re.match(
                r"(?:Complete|Season|Full)?[\]\[,.+\- ]*(?:Complete|Season|Full)?\Z",
                extra,
                re.IGNORECASE,
            ):
                filtered.append(extra)
        return filtered



#!/usr/bin/env python

import pkgutil
import sys

# Regex in python 2 is very slow so we check if the faster 'regex' library is available.
faster_regex = pkgutil.find_loader("regex")
if faster_regex is not None and sys.version_info[0] < 3:
    re = faster_regex.load_module("regex")
else:
    re = pkgutil.find_loader("re").load_module("re")

from .parse import PTN

__author__ = "Giorgio Momigliano"
__email__ = "gmomigliano@protonmail.com"
__version__ = "2.5"
__license__ = "MIT"


# TODO change coherent_types default to True in 3.0
def parse(name, standardise=True, coherent_types=False):
    return PTN().parse(name, standardise, coherent_types)



#!/usr/bin/env python

# Helper functions and constants for patterns.py

delimiters = "[\.\s\-\+_\/(),]"

langs = [
    ("rus(?:sian)?", "Russian"),
    ("(?:True)?fre?(?:nch)?", "French"),
    ("(?:nu)?ita(?:liano?)?", "Italian"),
    ("castellano|spa(?:nish)?|esp?", "Spanish"),
    ("swedish", "Swedish"),
    ("dk|dan(?:ish)?", "Danish"),
    ("ger(?:man)?|deu(?:tsch)?", "German"),
    ("nordic", "Nordic"),
    ("exyu", "ExYu"),
    ("chs|chi(?:nese)?", "Chinese"),
    ("hin(?:di)?", "Hindi"),
    ("polish|poland", "Polish"),
    ("mandarin", "Mandarin"),
    ("kor(?:ean)?", "Korean"),
    ("bengali|bangla", "Bengali"),
    ("kannada", "Kannada"),
    ("tam(?:il)?", "Tamil"),
    ("tel(?:ugu)?", "Telugu"),
    ("marathi", "Marathi"),
    ("mal(?:ayalam)?", "Malayalam"),
    ("japanese|ja?p", "Japanese"),
    ("interslavic", "Interslavic"),
    ("ara(?:bic)?", "Arabic"),
    ("urdu", "Urdu"),
    ("punjabi", "Punjabi"),
    ("portuguese", "Portuguese"),
    ("albanian?", "Albanian"),
    ("egypt(?:ian)?", "Egyptian"),
    ("en?(?:g(?:lish)?)?", "English"),  # Must be at end, matches just an 'e'
]

genres = [
    ("Sci-?Fi", "Sci-Fi"),
    ("Drama", "Drama"),
    ("Comedy", "Comedy"),
    ("West(?:\.|ern)?", "Western"),
    ("Action", "Action"),
    ("Adventure", "Adventure"),
    ("Thriller", "Thriller"),
]

# Some titles just can't be parsed without breaking everything else, so here
# are known those known exceptions. They are executed when the parsed_title and
# incorrect_parse match within a .parse() dict, removing the latter, and replacing
# the former with actual_title.
exceptions = [
    {"parsed_title": "", "incorrect_parse": ("year", 1983), "actual_title": "1983"},
    {
        "parsed_title": "Marvel's Agents of S H I E L D",
        "incorrect_parse": ("title", "Marvel's Agents of S H I E L D"),
        "actual_title": "Marvel's Agents of S.H.I.E.L.D.",
    },
    {
        "parsed_title": "Marvels Agents of S H I E L D",
        "incorrect_parse": ("title", "Marvels Agents of S H I E L D"),
        "actual_title": "Marvel's Agents of S.H.I.E.L.D.",
    },
    {
        "parsed_title": "Magnum P I",
        "incorrect_parse": ("title", "Magnum P I"),
        "actual_title": "Magnum P.I.",
    },
]

# Patterns that should only try to be matched after the 'title delimiter', either a year
# or a season. So if we have a language in the title it won't cause issues by getting matched.
# Empty list indicates to always do so, as opposed to matching specific regexes.
patterns_ignore_title = {
    "language": [],
    "audio": ["LiNE"],
    "network": ["Hallmark"],
    "untouched": [],
    "internal": [],
    "limited": [],
    "proper": [],
    "extended": [],
}


channels = [(1, 0), (2, 0), (5, 1), (6, 1), (7, 1)]


# Return tuple with regexes for audio name with appended channel types, and without any channels
def get_channel_audio_options(patterns_with_names):
    options = []
    for (audio_pattern, name) in patterns_with_names:
        for (speakers, subwoofers) in channels:
            options.append(
                (
                    "((?:{}){}*{}[. \-]?{}(?:ch)?)".format(
                        audio_pattern, delimiters, speakers, subwoofers
                    ),
                    "{} {}.{}".format(name, speakers, subwoofers),
                )
            )
        options.append(
            ("({})".format(audio_pattern), name)
        )  # After for loop, would match first

    return options


def prefix_pattern_with(prefixes, pattern_options, between="", optional=False):
    if optional:
        optional_char = "?"
    else:
        optional_char = ""
    options = []
    if not isinstance(prefixes, list):
        prefixes = [prefixes]
    for prefix in prefixes:
        if not isinstance(pattern_options, list):
            pattern_options = [pattern_options]
        for pattern_option in pattern_options:
            if isinstance(pattern_option, str):
                options.append(
                    "(?:{}){}(?:{})?({})".format(
                        prefix, optional_char, between, pattern_option
                    )
                )
            else:
                options.append(
                    (
                        "(?:{}){}(?:{})?({})".format(
                            prefix, optional_char, between, pattern_option[0]
                        ),
                    )
                    + pattern_option[1:]
                )

    return options


def suffix_pattern_with(suffixes, pattern_options, between="", optional=False):
    if optional:
        optional_char = "?"
    else:
        optional_char = ""
    options = []
    if not isinstance(suffixes, list):
        suffixes = [suffixes]
    for suffix in suffixes:
        if not isinstance(pattern_options, list):
            pattern_options = [pattern_options]
        for pattern_option in pattern_options:
            if isinstance(pattern_option, tuple):
                options.append(
                    (
                        "({})(?:{})?(?:{}){}".format(
                            pattern_option[0], between, suffix, optional_char
                        ),
                    )
                    + pattern_option[1:]
                )
            else:
                options.append(
                    "({})(?:{})?(?:{}){}".format(
                        pattern_option, between, suffix, optional_char
                    )
                )

    return options


# Link a regex-tuple list into a single regex (to be able to use elsewhere while
# maintaining standardisation functionality).
def link_patterns(pattern_options):
    if not isinstance(pattern_options, list):
        return pattern_options
    return (
        "(?:"
        + "|".join(
            [
                pattern_option[0]
                if isinstance(pattern_option, tuple)
                else pattern_option
                for pattern_option in pattern_options
            ]
        )
        + ")"
    )



#!/usr/bin/env python

# Patterns are either just a regex, or a tuple (or list of tuples) that contain the regex
# to match, (optional) what it should be replaced with (None if to not replace), and
# (optional) a string function's name to transform the value after everything (None if
# to do nothing). The transform can also be a tuple (or list of tuples) with function names
# and list of arguments.
# The list of regexes all get matched, but only the first gets added to the returning info,
# the rest are just matched to be removed from `excess`.

from .extras import (
    delimiters,
    genres,
    get_channel_audio_options,
    langs,
    link_patterns,
    suffix_pattern_with,
)

season_range_pattern = (
    "(?:Complete"
    + delimiters
    + "*)?"
    + delimiters
    + "*(?:s(?:easons?)?)"
    + delimiters
    + "*(?:s?[0-9]{1,2}[\s]*(?:(?:\-|(?:\s*to\s*))[\s]*s?[0-9]{1,2})+)(?:"
    + delimiters
    + "*Complete)?"
)

year_pattern = "(?:19[0-9]|20[0-2])[0-9]"
month_pattern = "0[1-9]|1[0-2]"
day_pattern = "[0-2][0-9]|3[01]"

episode_name_pattern = (
    "((?:[Pp](?:ar)?t"
    + delimiters
    + "*[0-9]|(?:[A-Za-z]|[0-9])[a-z]*(?:"
    + delimiters
    + "|$))+)"
)
pre_website_encoder_pattern = r"[^\s\.\[\]\-\(\)]+\)\s{0,2}\[[^\s\-]+\]|[^\s\.\[\]\-\(\)]+\s{0,2}(?:-\s)?[^\s\.\[\]\-]+$"

# Forces an order to go by the regexes, as we want this to be deterministic (different
# orders can generate different matchings). e.g. "doctor_who_2005..." in input.json
patterns_ordered = [
    "season",
    "episode",
    "year",
    "month",
    "day",
    "resolution",
    "quality",
    "codec",
    "audio",
    "region",
    "extended",
    "hardcoded",
    "proper",
    "repack",
    "filetype",
    "widescreen",
    "site",
    "documentary",
    "language",
    "subtitles",
    "sbs",
    "unrated",
    "size",
    "bitDepth",
    "3d",
    "internal",
    "readnfo",
    "network",
    "fps",
    "hdr",
    "limited",
    "remastered",
    "directorsCut",
    "upscaled",
    "untouched",
    "remux",
    "internationalCut",
    "genre",
]

patterns = {}
patterns["episode"] = [
    "(?<![a-z])(?:e|ep)(?:[0-9]{1,2}(?:-?(?:e|ep)?(?:[0-9]{1,2}))?)(?![0-9])",
    # Very specific as it could match too liberally
    "\s\-\s\d{1,3}\s",
    r"\b[0-9]{1,2}x([0-9]{2})\b",
    r"\bepisod(?:e|io)" + delimiters + r"\d{1,2}\b",
]
# If adding season patterns, remember to look at episode, as it uses the last few!
patterns["season"] = [
    "\ss?(\d{1,2})\s\-\s\d{1,2}\s",  # Avoids matching some anime releases season and episode as a season range
    r"\b" + season_range_pattern + r"\b",  # Describes season ranges
    r"(?:s\d{1,2}[.+\s]*){2,}\b",  # for S01.S02.etc. patterns
    # Describes season, optionally with complete or episode
    r"\b(?:Complete"
    + delimiters
    + ")?s([0-9]{1,2})"
    + link_patterns(patterns["episode"])
    + r"?\b",
    r"\b([0-9]{1,2})x[0-9]{2}\b",  # Describes 5x02, 12x15 type descriptions
    "[0-9]{1,2}(?:st|nd|rd|th)" + delimiters + "season",
    "Series" + delimiters + "\d{1,2}",
    r"\b(?:Complete"
    + delimiters
    + r")?Season[\. -][0-9]{1,2}\b",  # Describes Season.15 type descriptions
]
# The first 4 season regexes won't have 'Part' in them.
patterns["episode"] += [
    link_patterns(patterns["season"][5:])
    + delimiters
    + "*P(?:ar)?t"
    + delimiters
    + "*(\d{1,3})"
]
patterns["year"] = year_pattern
patterns["month"] = "(?:{year}){d}({month}){d}(?:{day})".format(
    d=delimiters, year=year_pattern, month=month_pattern, day=day_pattern
)
patterns["day"] = "(?:{year}){d}(?:{month}){d}({day})".format(
    d=delimiters, year=year_pattern, month=month_pattern, day=day_pattern
)
patterns["resolution"] = [
    ("([0-9]{3,4}(?:p|i))", None, "lower"),
    ("(1280x720p?)", "720p"),
    ("FHD|1920x1080p?", "1080p"),
    ("UHD", "UHD"),
    ("HD", "HD"),
    ("4K", "4K"),
]
patterns["quality"] = [
    ("WEB[ -\.]?DL(?:Rip|Mux)?|HDRip", "WEB-DL"),
    # Match WEB-DL's first as they can show up with others.
    ("WEB[ -]?Cap", "WEBCap"),
    ("W[EB]B[ -]?(?:Rip)|WEB", "WEBRip"),
    ("(?:HD)?CAM(?:-?Rip)?", "Cam"),
    ("(?:HD)?TS|TELESYNC|PDVD|PreDVDRip", "Telesync"),
    ("WP|WORKPRINT", "Workprint"),
    ("(?:HD)?TC|TELECINE", "Telecine"),
    ("(?:DVD)?SCR(?:EENER)?|BDSCR", "Screener"),
    ("DDC", "Digital Distribution Copy"),
    ("DVD-?(?:Rip|Mux)", "DVD-Rip"),
    ("DVDR|DVD-Full|Full-rip", "DVD-R"),
    ("PDTV|DVBRip", "PDTV"),
    ("DSR(?:ip)?|SATRip|DTHRip", "DSRip"),
    ("AHDTV(?:Mux)?", "AHDTV"),
    ("HDTV(?:Rip)?", "HDTV"),
    ("D?TVRip|DVBRip", "TVRip"),
    ("VODR(?:ip)?", "VODRip"),
    ("HD-Rip", "HD-Rip"),
    ("Blu-?Ray{d}Rip|BDR(?:ip)?".format(d=delimiters), "BDRip"),
    ("Blu-?Ray|(?:US|JP)?BD(?:remux)?", "Blu-ray"),
    ("BR-?Rip", "BRRip"),
    ("HDDVD", "HD DVD"),
    # Match this last as it can show up with others.
    ("PPV(?:Rip)?", "Pay-Per-View Rip"),
]
patterns["network"] = [
    ("ATVP", "Apple TV+"),
    ("AMZN|Amazon", "Amazon Studios"),
    ("NF|Netflix", "Netflix"),
    ("NICK", "Nickelodeon"),
    ("RED", "YouTube Premium"),
    ("DSNY?P", "Disney Plus"),
    ("DSNY", "DisneyNOW"),
    ("HMAX", "HBO Max"),
    ("HBO", "HBO"),
    ("HULU", "Hulu Networks"),
    ("MS?NBC", "MSNBC"),
    ("DCU", "DC Universe"),
    ("ID", "Investigation Discovery"),
    ("iT", "iTunes"),
    ("AS", "Adult Swim"),
    ("CRAV", "Crave"),
    ("CC", "Comedy Central"),
    ("SESO", "Seeso"),
    ("VRV", "VRV"),
    ("PCOK", "Peacock"),
    ("CBS", "CBS"),
    ("iP", "BBC iPlayer"),
    ("NBC", "NBC"),
    ("AMC", "AMC"),
    ("PBS", "PBS"),
    ("STAN", "Stan."),
    ("RTE", "RTE Player"),
    ("CR", "Crunchyroll"),
    ("ANPL", "Animal Planet Live"),
    ("DTV", "DirecTV Stream"),
    ("VICE", "VICE"),
]
patterns["network"] = suffix_pattern_with(
    link_patterns(patterns["quality"]), patterns["network"], delimiters
)
# Not all networks always show up just before the quality, so if they're unlikely to clash,
# they should be added here.
patterns["network"] += [
    ("BBC", "BBC"),
    ("Hoichoi", "Hoichoi"),
    ("Zee5", "ZEE5"),
    ("Hallmark", "Hallmark"),
    ("Sony\s?LIV", "SONY LIV"),
]
patterns["codec"] = [
    ("xvid", "Xvid"),
    ("av1", "AV1"),
    ("[hx]{d}?264".format(d=delimiters), "H.264"),
    ("AVC", "H.264"),
    ("HEVC(?:{d}Main{d}?10P?)".format(d=delimiters), "H.265 Main 10"),
    (
        "[hx]{d}?265".format(d=delimiters),
        "H.265",
    ),  # Separate from HEVC so if both are present, it won't pollute excess.
    ("HEVC", "H.265"),
    ("[h]{d}?263".format(d=delimiters), "H.263"),
    ("VC-1", "VC-1"),
]
patterns["audio"] = get_channel_audio_options(
    [
        ("TrueHD", "Dolby TrueHD"),
        ("Atmos", "Dolby Atmos"),
        ("DD-EX", "Dolby Digital EX"),
        ("DD|AC-?3|DolbyD", "Dolby Digital"),
        ("DDP|E-?AC-?3|EC-3", "Dolby Digital Plus"),
        (
            "DTS{d}?HD(?:{d}?(?:MA|Masters?(?:{d}Audio)?))".format(d=delimiters),
            "DTS-HD MA",
        ),
        ("DTSMA", "DTS-HD MA"),
        ("DTS{d}?HD".format(d=delimiters), "DTS-HD"),
        ("DTS", "DTS"),
        ("AAC[ \.\-]LC", "AAC-LC"),
        ("AAC", "AAC"),
        ("Dual{d}Audios?".format(d=delimiters), "Dual"),
        ("FLAC", "FLAC"),
        ("OGG", "OGG"),
    ]
) + [
    ("7.1(?:{d}?ch(?:annel)?(?:{d}?Audio)?)?".format(d=delimiters), "7.1"),
    ("5.1(?:{d}?ch(?:annel)?(?:{d}?Audio)?)?".format(d=delimiters), "5.1"),
    ("MP3", None, "upper"),
    ("2.0(?:{d}?ch(?:annel)?(?:{d}?Audio)?)?|2CH|stereo".format(d=delimiters), "Dual"),
    ("1{d}?Ch(?:annel)?(?:{d}?Audio)?".format(d=delimiters), "Mono"),
    ("(?:Original|Org)" + delimiters + "Aud(?:io)?", "Original"),
    ("LiNE", "LiNE"),
]
patterns["region"] = ("R[0-9]", None, "upper")
patterns["extended"] = "(EXTENDED(:?.CUT)?)"
patterns["hardcoded"] = "HC"
patterns["proper"] = "PROPER"
patterns["repack"] = "REPACK"
patterns["fps"] = "([1-9][0-9]{1,2})" + delimiters + "*fps"
patterns["filetype"] = [(r"\.?(MKV|AVI|(?:SRT|SUB|SSA)$)", None, "upper"), ("MP-?4", "MP4")]
patterns["widescreen"] = "WS"
patterns["site"] = r"^(\[ ?([^\]]+?) ?\])"

lang_list_pattern = (
    r"\b(?:"
    + link_patterns(langs)
    + "(?:"
    + delimiters
    + "+(?:dub(?:bed)?|"
    + link_patterns(patterns["audio"])
    + "))?"
    + "(?:"
    + delimiters
    + r"+|\b))"
)
subs_list_pattern = r"\b(?:" + link_patterns(langs) + delimiters + "*)"

patterns["subtitles"] = [
    "sub(?:title|bed)?s?{d}*{langs}+".format(d=delimiters, langs=subs_list_pattern),
    "(?:soft{d}*)?{langs}+(?:(?:m(?:ulti(?:ple)?)?{d}*)?sub(?:title|bed)?s?)".format(
        d=delimiters, langs=subs_list_pattern
    ),
    # Need a pattern just for subs, and can't just make above regexes * over + as we want
    # just 'subs' to match last.
    "(?:m(?:ulti(?:ple)?)?{d}*)sub(?:title|bed)?s?".format(d=delimiters),
    "(?:m(?:ulti(?:ple)?)?[\.\s\-\+_\/]*)?sub(?:title|bed)?s?{d}*".format(d=delimiters),
]
# Language takes precedence over subs when ambiguous - if we have a lang match, and
# then a subtitles match starting with subs, the first langs are languages, and the
# rest will be left as subtitles. Otherwise, don't match if there are subtitles matches
# after the langs.
patterns["language"] = [
    "("
    + lang_list_pattern
    + "+)(?:"
    + delimiters
    + "*"
    + patterns["subtitles"][0]
    + ")",
    "("
    + lang_list_pattern
    + "+)"
    + ""
    + "(?!"
    + delimiters
    + "*"
    + link_patterns(patterns["subtitles"])
    + ")",
    "("
    + lang_list_pattern
    + "+)(?:"
    + delimiters
    + "*"
    + patterns["subtitles"][2]
    + ")",
]
patterns["sbs"] = [("Half-SBS", "Half SBS"), ("SBS", None, "upper")]
patterns["unrated"] = "UNRATED"
patterns["size"] = (
    "\d+(?:\.\d+)?\s?(?:GB|MB)",
    None,
    [("upper", []), ("replace", [" ", ""])],
)
patterns["bitDepth"] = "(8|10)-?bits?"
patterns["3d"] = "3D"
patterns["internal"] = "iNTERNAL"
patterns["readnfo"] = "READNFO"
patterns["hdr"] = "HDR(?:10)?"
patterns["documentary"] = "DOCU(?:menta?ry)?"
patterns["limited"] = "LIMITED"
patterns["remastered"] = "REMASTERED"
patterns["directorsCut"] = "DC|Director'?s.?Cut"
patterns["upscaled"] = "(?:AI{d}*)?upscaled?".format(d=delimiters)
patterns["untouched"] = "untouched"
patterns["remux"] = "REMUX"
patterns["internationalCut"] = "International{d}Cut".format(d=delimiters)
# Spaces are only allowed before the genre list if after a word boundary or punctuation
patterns["genre"] = (
    r"\b\s*[\(\-\]]+\s*((?:" + link_patterns(genres) + delimiters + r"?)+)\b"
)

types = {
    "season": "integer",
    "episode": "integer",
    "bitDepth": "integer",
    "year": "integer",
    "month": "integer",
    "day": "integer",
    "fps": "integer",
    "extended": "boolean",
    "hardcoded": "boolean",
    "proper": "boolean",
    "repack": "boolean",
    "widescreen": "boolean",
    "unrated": "boolean",
    "3d": "boolean",
    "internal": "boolean",
    "readnfo": "boolean",
    "documentary": "boolean",
    "hdr": "boolean",
    "limited": "boolean",
    "remastered": "boolean",
    "directorsCut": "boolean",
    "upscaled": "boolean",
    "untouched": "boolean",
    "remux": "boolean",
    "internationalCut": "boolean",
}



import codecs
import os

from setuptools import setup

readme_path = os.path.join(os.path.dirname(__file__), "README.md")

with codecs.open(readme_path, mode="r", encoding="utf-8") as f:
    description = f.read()

setup(
    name="parse-torrent-title",
    version=__import__("PTN").__version__,
    author=__import__("PTN").__author__,
    author_email=__import__("PTN").__email__,
    license=__import__("PTN").__license__,
    url="https://github.com/platelminto/parse-torrent-title",
    description="Extract media information from torrent-like filename",
    long_description=description,
    long_description_content_type="text/markdown",
    packages=["PTN"],
    keywords=(
        "parse parser torrent torrents name names proper rename "
        "movie movies tv show shows series extract find quality "
        "group codec audio resolution title season episode year "
        "information filename filenames file files meaningful"
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Topic :: Text Processing",
    ],
)
